"""
Legal Disclaimer

(c) 2018 Amazon Web Services, Inc. or its affiliates. All Rights Reserved.
This AWS Content is provided subject to the terms of the AWS Customer
Agreement available at https://aws.amazon.com/agreement/ or other written
agreement between Customer and Amazon Web Services, Inc.

Information Block

All data stored in Redshift must be encrypted using KMS.
API Triggers: CreateCluster, RestoreFromClusterSnapshot
Services: Amazon Cloudwatch Events (trigger), Amazon Simple Notification
Service (Amazon SNS),Amazon Redshift Cluster.
Python 3.6 - AWS Lambda - Last Modified 05/09/2018
"""


import os
import json
import boto3
import logging
from botocore.exceptions import ClientError

OUTBOUND_TOPIC_ARN = os.environ["outbound_topic_arn"]


def lambda_handler(event, context):
    """
    Main Function
    Check Redshift cluster configuration to validate data at rest encrypted using KMS.
    """
    setup_logging()
    log.info(json.dumps(event))
    if unexpected_response_handler(event):
        return 'Unexpected Response from AWS'
    try:
        if event["detail"]["eventName"] in ["CreateCluster"] or ["RestoreFromClusterSnapshot"]:
            if not is_encrypted(event):
                subject = "Violation - Amazon Redshift Cluster is NON-COMPLIANT"
                message = create_non_compliance_message("Violation: No encryption key associated", event)
                send_violation(OUTBOUND_TOPIC_ARN, message,subject, event, context)
            else:
                log.info ("No Violation - Customer provided KMS key assoicated")
    except KeyError as e:
        log.error(e)

def is_encrypted(event):
    """
    Check if Redshift associated with KMS encryption key.

    This function will check if the newly created or Restored Amazon Redshift is associated with a KMS key.
    """
    try:
        if event["detail"]["responseElements"] is None:
            log.warning('Unexpected response element. Triggering violation.')
            return False
        if "kmsKeyId" in event["detail"]["requestParameters"]:
            kmsKeyARN = (event["detail"]["requestParameters"]["kmsKeyId"]).split(" ")
            kmsKeyID = kmsKeyARN[0]
            return kmsKeyID
        elif "kmsKeyId" in event["detail"]["responseElements"]:
            kmsKeyARN = (event["detail"]["responseElements"]["kmsKeyId"]).split(" ")
            kmsKeyID = kmsKeyARN[0]
            return kmsKeyID
        else:
            return False
    except KeyError as e:
        log.error(e)


def describe_kms(kmsKeyID):
    """
    Makes describe_key API

    This function call describe_key API and return the output
    """
    client = boto3.client('kms')
    try:
        response = client.describe_key(KeyId=kmsKeyID)
        keyManager = response["KeyMetadata"]["KeyManager"]
        return keyManager
    except ClientError as err:
        log.info(err)
        return False
  
def create_non_compliance_message(violation, event):
    """
    Creates non compliance message

    This function will create non-compliance message based on violation report
    """
    message = "Violation - The following Amazon Redshift Cluster was detected to be in violation of the below security compliance \n\n"
    message += "Amazon Redshift Cluster ID : " + \
        event["detail"]["requestParameters"]["clusterIdentifier"] + "\n"
    message += violation + "\n"
    return message


def send_violation(outbound_topic_arn, message, subject, event, context):
    """
    Build A Message

    Creates a message with a first line stating what this violation is checking
    Creates more lines with identifying information about which specific violations were checked
    This is based upon whether or not the specific violation triggered a "Violation" variable
    Then it adds general information regarding the violation
    """
    findsnsregion = outbound_topic_arn.split(":")
    snsregion = findsnsregion[3]
    log.warning('Sending Violation')
    message += '\nAccount: ' + event["account"] + "\n"
    message += "Region: " + event["detail"]["awsRegion"] + "\n"
    message += "\n\n"
    message += "This notification was generated by the Lambda function " + \
        context.invoked_function_arn
    sendclient = boto3.client('sns', region_name=snsregion)
    try:
        response = sendclient.publish(
            TopicArn=outbound_topic_arn,
            Message=message,
            Subject=subject
            )
    except ClientError as err:
        log.info (err)
        return False

def setup_logging():
    """
    Logging Function.

    Creates a global log object and sets its level.
    """
    global log
    log = logging.getLogger()
    log_levels = {'INFO': 20, 'WARNING': 30, 'ERROR': 40}

    if 'logging_level' in os.environ:
        log_level = os.environ['logging_level'].upper()
        if log_level in log_levels:
            log.setLevel(log_levels[log_level])
        else:
            log.setLevel(log_levels['ERROR'])
            log.error("The logging_level environment variable is not set to INFO, WARNING, or \
                        ERROR.  The log level is set to ERROR")
    else:
        log.setLevel(log_levels['ERROR'])
    log.info('Logging setup complete - set to log level ' + str(log.getEffectiveLevel()))

def unexpected_response_handler(event):
    """
    Handles unexpected events.

    This function handles unexpected responses in CloudTrail
    such as limit exceptions or error message response from AWS.
    """
    try:
        if event["detail"]["responseElements"] is None:
            log.info('Null responseElements detected.')
            return True
        elif 'errorCode' in event["detail"]:
            log.warning('errorCode detected, request discarded.')
            return True
        elif 'errorCode' in event["detail"]["responseElements"]:
            log.warning('errorCode detected, request discarded.')
            return True
        else:
            return False
    except KeyError as err:
        log.error(err)